/**
 * BlinkMComm -- Simple Processing 'library' to talk to BlinkM 
 *               (via an Arduino programmed with BlinkMCommunicator)
 *
 *
 * Copyleft 2007, Tod E. Kurt, ThingM, http://thingm.com/
 * 
 */

import processing.serial.*;

//import javax.swing.progressbar;

public class BlinkMComm {
  public final boolean fakeIt = false;

  boolean isConnected = false;

  public byte blinkm_addr = 0x09;
  public String portName = null;
  public final int portSpeed = 19200;

  Serial port;

  // mapping of duration to ticks      (must be same length as 'durations')
  public byte[] durTicks   = { (byte)   1, (byte) 10, (byte) 100 };
  // mapping of duration to fadespeeds (must be same length as 'durations')
  public byte[] fadeSpeeds = { (byte) 100, (byte) 10, (byte)   5 };
    
  Color cBlk = new Color(0,0,0);
  Color lastColor;

  // Return a list of potential ports
  // they should be ordered by best to worst (but are not right now)
  // this can't be static as a .pde, sigh.
  public String[] listPorts() {
    String[] a = Serial.list();
    String osname = System.getProperty("os.name");
    if( osname.toLowerCase().startsWith("windows") ) {
      // reverse list because Arduino is almost always highest COM port
      for(int i=0;i<a.length/2;i++){
        String t = a[i]; a[i] = a[a.length-(1+i)]; a[a.length-(1+i)] = t;
      }
      //for(int left=0, int right=list.length-1; left<right; left++, right--) {
      //  // exchange the first and last
      //  String tmp = list[left]; list[left] = list[right]; list[right] = tmp;
      //}
    }
    return a;
  }

  public BlinkMComm() {

  }

  /**
   * Connect to the given port
   * Can optionally take a PApplet (the sketch) to get serialEvents()
   * but this is not recommended
   *
   */
  public void connect( PApplet p, String portname ) throws Exception {
    l.debug("BlinkMComm.connect: portname:"+portname);
    try {
      if(port != null)
        port.stop(); 
      port = new Serial(p, portname, portSpeed);
      isConnected = true;
      portName = portname;
    }
    catch (Exception e) {
      isConnected = false;
      portName = null;
      port = null;
      throw e;
    }
  }

  // disconnect but remember the name
  public void disconnect() {
    if( port!=null )
      port.stop();
    isConnected = false;
  }

  // verifies connection is good
  public boolean checkConnection() {
    // FIXME: add echo check
    return true;
  }

  public boolean isConnected() {
    return isConnected; // FIXME: this is kinda lame
  }

  // uses global var 'durations'
  public byte getDurTicks(int loopduration) {
    for( int i=0; i<durations.length; i++ ) {
      if( durations[i] == loopduration )
        return durTicks[i];
    }
    return durTicks[0]; // failsafe
  }
  // this is so lame
  public byte getFadeSpeed(int loopduration) {
    for( int i=0; i<durations.length; i++ ) {
      if( durations[i] == loopduration )
        return fadeSpeeds[i];
    }
    return fadeSpeeds[0]; // failsafe
  }

  /**
   * Burn a list of colors to a BlinkM
   * @param colorlist an ArrayList of the Colors to burn (java Color objs)
   * @param nullColor a color in the list that should be treated as nothing
   * @param duration  how long the entire list should last for, in seconds
   * @param loop      should the list be looped or not
   * @param progressbar if not-null, will update a progress bar
   */
  public void burn(ArrayList colorlist, Color nullColor, 
                   int duration, boolean loop, 
                   JProgressBar progressbar) {

    byte[] cmd = new byte[8];
    byte fadespeed = getFadeSpeed(duration);
    byte durticks = getDurTicks(duration);
    Color c;

    l.debug("BlinkMComm.burn: durticks:"+durticks+" fadespeed:"+fadespeed);

    // build up the byte array to send
    Iterator iter = colorlist.iterator();
    int i=0;
    while( iter.hasNext() ) {
      //l.debug("BlinkMComm.burn: writing script line "+i);
      c = (Color) iter.next();
      if( c == nullColor )
        c = cBlk;
      cmd[0] = (byte)'W';          // "Write Script Line" command
      cmd[1] = (byte)0;            // script id (0==eeprom)
      cmd[2] = (byte)i;            // script line number
      cmd[3] = (byte)durticks;     // duration in ticks
      cmd[4] = (byte)'c';          // fade to rgb color command
      cmd[5] = (byte)c.getRed();   // cmd arg1
      cmd[6] = (byte)c.getGreen(); // cmd arg2
      cmd[7] = (byte)c.getBlue();  // cmd arg3
      sendCommand( blinkm_addr, cmd );
      if( progressbar !=null) progressbar.setValue(i);  // hack
      i++;
      pause(50);
    }
    // set script length   cmd   id         length         reps
    byte[] cmdsetlength = { 'L', 0, (byte)colorlist.size(), 0 };
    sendCommand( blinkm_addr, cmdsetlength );
    pause(30);

    // set boot params   cmd mode 
    //byte cmdsetboot[] = { 'B', 1, 0, 0, fadespeed, 0 };
    //sendCommand( blinkm_addr, cmdsetboot );
    //pause(30);

    // set fade speed
    byte[] cmdsetfade = { 'f', fadespeed };
    sendCommand( blinkm_addr, cmdsetfade );
    pause(30);

    // and cause the script to be played (infinitely)
    byte[] cmdplay = { 'p', 0, 0, 0 };
    sendCommand( blinkm_addr, cmdplay );
    pause(30);
  }

  // prepare blinkm for playing preview scripts
  public void prepareForPreview(int loopduration) {
    byte fadespeed = getFadeSpeed(loopduration);
    l.debug("BlinkmComm.prepareForPreview: fadespeed:"+fadespeed);
    byte[] cmdstop    = {'o'};
    byte[] cmdsetfade = {'f', fadespeed};
    if( isConnected() ) {
      sendCommand( blinkm_addr, cmdstop );
      pause(30);
      sendCommand( blinkm_addr, cmdsetfade );
      pause(30);
    }
  }

  // stops any playing script
  public void stopPlayingScript() {
    l.debug("BlinkmComm.stopPlayingScript");
    byte[] cmd = {'o'};
    if( isConnected() ) 
      sendCommand( blinkm_addr, cmd );
  }


  /**
   *
   */
  public void sendColor( Color aColor, Color nullColor, int duration ) {
    if( aColor.equals( lastColor ) )   // don't clog the pipes!
      return;

    Color c = (aColor == nullColor) ? cBlk : aColor;
    byte[] cmd={'c',(byte)c.getRed(),(byte)c.getGreen(),(byte)c.getBlue()};
    if( isConnected() ) 
      sendCommand( blinkm_addr, cmd );
    pause(10); // FIXME: hack
    lastColor = aColor;
  }

  /**
   * Send an I2C command to addr, via the BlinkMCommander Arduino sketch
   * Byte array must be correct length
   */
  public synchronized void sendCommand( byte addr, byte[] cmd ) {
    if( fakeIt ) return;        // just pretend
    l.debug("BlinkMComm.sendCommand(): "+(char)cmd[0]+", "+cmd.length);
    byte cmdfull[] = new byte[4+cmd.length];
    cmdfull[0] = 0x01;
    cmdfull[1] = addr;
    cmdfull[2] = (byte)cmd.length;
    cmdfull[3] = 0x00;
    for( int i=0; i<cmd.length; i++)
      cmdfull[4+i] = cmd[i];
    port.write(cmdfull);
  }

  /**
   * A simple delay
   */
  public void pause( int millis ) {
    try { Thread.sleep(millis); } catch(Exception e) { }
  }

}

